{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "[Day 11](https://adventofcode.com/2019/day/11) puzzle input is an IntCode program which is supposed to control a little robot that paints the surface of a ship (mine is [here](puzzle_input/day11.txt)). Part 1 involves running the program and keeping track of the number of panels it paints. Part 2 involves running the program over a panel coloured white, keeping track of what the robot paints - which should spell out a registration number.\n",
    "\n",
    "Note: this is one of the ones I disliked because a few instructions were rather vague and weirdly worded, like this:\n",
    "\n",
    "> The panel under the robot (not visible here because a ^ is shown instead) is also black, and so any input instructions at this point should be provided `0`.\n",
    "\n",
    "This \"provided `N`\" appears a few times in the text."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 1,
   "metadata": {},
   "outputs": [],
   "source": [
    "opcode_add = 1\n",
    "opcode_mul = 2\n",
    "opcode_read = 3\n",
    "opcode_write = 4\n",
    "opcode_jump_true = 5\n",
    "opcode_jump_false = 6\n",
    "opcode_lt = 7\n",
    "opcode_eq = 8\n",
    "opcode_rebase = 9\n",
    "\n",
    "opcode_terminate = 99\n",
    "\n",
    "mode_position = 0\n",
    "mode_immediate = 1\n",
    "mode_relative = 2\n",
    "\n",
    "class IntCodeCpu:\n",
    "    def __init__(self, memory_image):\n",
    "        self.memory = [ x for x in memory_image ] # copy memory image, in case it's reused\n",
    "        self.stalled = True\n",
    "        self.input_buffer = None\n",
    "        self.output_buffer = None\n",
    "        self.pc = 0\n",
    "        self.initialise_opcodes()\n",
    "        self.offset = 0\n",
    "        self.done = False\n",
    "        \n",
    "    def start(self, input_buffer, output_buffer, noun=None, verb=None):\n",
    "        self.input_buffer = input_buffer\n",
    "        self.output_buffer = output_buffer\n",
    "\n",
    "        if noun:\n",
    "            self.memory[1] = noun\n",
    "        if verb:\n",
    "            self.memory[2] = verb\n",
    "           \n",
    "        return self.run()\n",
    "            \n",
    "            \n",
    "    def run(self):\n",
    "        instr = self.memory[self.pc]\n",
    "        self.stalled = False\n",
    "\n",
    "        while int(instr) != opcode_terminate and not self.stalled:\n",
    "            (op, modes) = self.decode_instr(instr)\n",
    "            self.pc = op(modes)\n",
    "            instr = self.memory[self.pc]\n",
    "\n",
    "        return self.memory[0]\n",
    "\n",
    "    #-HELPERS-----------------------------\n",
    "    def try_pop_mode(self, modes):\n",
    "        if len(modes) == 0:\n",
    "            return 0\n",
    "    \n",
    "        return modes.pop()\n",
    "    \n",
    "    def resize_memory(self, target_addr):\n",
    "        self.memory += ([0] * (1 + target_addr - len(self.memory)))\n",
    "    \n",
    "    #-DECODE-INSTRUCTIONS-----------------\n",
    "    def initialise_opcodes(self):\n",
    "        self.opcodes = {\n",
    "            opcode_add: self.op_add,\n",
    "            opcode_mul: self.op_mul,\n",
    "            opcode_read: self.op_read,\n",
    "            opcode_write: self.op_write,\n",
    "            opcode_jump_true: self.op_jump_true,\n",
    "            opcode_jump_false: self.op_jump_false,\n",
    "            opcode_lt: self.op_lt,\n",
    "            opcode_eq: self.op_eq,\n",
    "            opcode_rebase: self.op_rebase\n",
    "        }    \n",
    "    \n",
    "    def decode_instr(self, instr):\n",
    "        instr = str(instr)\n",
    "        opcode = self.decode_op(instr)\n",
    "        modes = self.decode_modes(instr)\n",
    "\n",
    "        if not (opcode in self.opcodes):\n",
    "            raise Exception(f\"Invalid opcode {opcode}\")\n",
    "\n",
    "        return (self.opcodes[opcode], modes)    \n",
    "    \n",
    "    def decode_op(self, instr):\n",
    "        if len(instr) > 2:\n",
    "            return int(instr[-2:])\n",
    "        return int(instr)\n",
    "    \n",
    "    def decode_modes(self, instr):\n",
    "        if len(instr) > 2:\n",
    "            return [ int(d) for d in instr[:-2]]\n",
    "        return []\n",
    "    \n",
    "    #-MICRO-OPS---------------------------\n",
    "    def uop_read(self, value, mode):\n",
    "        if mode == mode_position:\n",
    "            if value >= len(self.memory):\n",
    "                self.resize_memory(value)        \n",
    "\n",
    "            return int(self.memory[value])\n",
    "\n",
    "        elif mode == mode_relative:\n",
    "            if self.offset + value >= len(self.memory):\n",
    "                self.resize_memory(self.offset + value)     \n",
    "\n",
    "            return int(self.memory[self.offset + value])\n",
    "\n",
    "        elif mode == mode_immediate:\n",
    "            return int(value)\n",
    "\n",
    "        else:\n",
    "            raise Exception(\"UNKNOWN MODE\")\n",
    "\n",
    "    def uop_write(self, dst, value, mode):\n",
    "        if mode == mode_position:\n",
    "            if dst >= len(self.memory):\n",
    "                self.resize_memory(dst)\n",
    "            self.memory[dst] = value\n",
    "\n",
    "        elif mode == mode_relative:\n",
    "            if self.offset + dst >= len(self.memory):\n",
    "                self.resize_memory(self.offset + dst)     \n",
    "\n",
    "            self.memory[self.offset + dst] = value\n",
    "\n",
    "        elif mode == mode_immediate:\n",
    "            raise Exception(f\"cannot write {value} to literal {dst}\")\n",
    "\n",
    "\n",
    "    def uop_cond_jump(self, modes, cond):\n",
    "        param_mode = self.try_pop_mode(modes)\n",
    "        param_raw = int(self.memory[self.pc + 1])\n",
    "        param = self.uop_read(param_raw, param_mode)\n",
    "\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "        dest_raw = int(self.memory[self.pc + 2])\n",
    "        dest = self.uop_read(dest_raw, dest_mode)\n",
    "\n",
    "        if cond(param):\n",
    "            return dest\n",
    "\n",
    "        return self.pc + 3\n",
    "\n",
    "    def uop_cmp(self, modes, cmp):\n",
    "\n",
    "        param0_mode = self.try_pop_mode(modes)\n",
    "        param0_raw = int(self.memory[self.pc + 1])\n",
    "        param0 = self.uop_read(param0_raw, param0_mode)\n",
    "\n",
    "        param1_mode = self.try_pop_mode(modes)\n",
    "        param1_raw = int(self.memory[self.pc + 2])\n",
    "        param1 = self.uop_read(param1_raw, param1_mode)\n",
    "\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "        dest = int(self.memory[self.pc + 3])\n",
    "\n",
    "        if cmp(param0, param1):\n",
    "            self.uop_write(dest, 1, dest_mode)\n",
    "        else:\n",
    "            self.uop_write(dest, 0, dest_mode)\n",
    "\n",
    "        return self.pc + 4\n",
    "        \n",
    "    \n",
    "    #-OPCODES-----------------------------\n",
    "    def op_add(self, modes):\n",
    "        arg0_mode = self.try_pop_mode(modes)\n",
    "        arg1_mode = self.try_pop_mode(modes)\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "\n",
    "        arg0_raw = int(self.memory[self.pc + 1])\n",
    "        arg1_raw = int(self.memory[self.pc + 2])\n",
    "        dest = int(self.memory[self.pc + 3])\n",
    "\n",
    "        arg0 = self.uop_read(arg0_raw, arg0_mode)\n",
    "        arg1 = self.uop_read(arg1_raw, arg1_mode)\n",
    "\n",
    "        self.uop_write(dest, str(int(arg0) + int(arg1)), dest_mode)\n",
    "        return self.pc + 4\n",
    "\n",
    "    def op_mul(self, modes):\n",
    "        arg0_mode = self.try_pop_mode(modes)\n",
    "        arg1_mode = self.try_pop_mode(modes)\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "\n",
    "        arg0_raw = int(self.memory[self.pc + 1])\n",
    "        arg1_raw = int(self.memory[self.pc + 2])\n",
    "        dest = int(self.memory[self.pc + 3])\n",
    "\n",
    "        arg0 = self.uop_read(arg0_raw, arg0_mode)\n",
    "        arg1 = self.uop_read(arg1_raw, arg1_mode)\n",
    "\n",
    "        self.uop_write(dest, str(int(arg0) * int(arg1)), dest_mode)    \n",
    "        return self.pc + 4\n",
    "\n",
    "\n",
    "    def op_read(self, modes):\n",
    "        dest_mode = self.try_pop_mode(modes)\n",
    "        dest = int(self.memory[self.pc + 1])\n",
    "        \n",
    "        # if the input buffer is empty, we should \"stall\" and \n",
    "        # resume later\n",
    "        if not self.input_buffer:\n",
    "            self.stalled = True\n",
    "            return self.pc\n",
    "        \n",
    "        val = self.input_buffer.pop()\n",
    "\n",
    "        self.uop_write(dest, str(val), dest_mode)\n",
    "\n",
    "        return self.pc + 2\n",
    "\n",
    "    def op_write(self, modes):\n",
    "        src_mode = self.try_pop_mode(modes)\n",
    "        src_raw = int(self.memory[self.pc + 1])\n",
    "        src = self.uop_read(src_raw, src_mode)\n",
    "\n",
    "        self.output_buffer.append(src)\n",
    "\n",
    "        return self.pc + 2\n",
    "\n",
    "    def op_jump_true(self, modes):\n",
    "        return self.uop_cond_jump(modes, lambda x: x != 0)\n",
    "\n",
    "\n",
    "    def op_jump_false(self, modes):\n",
    "        return self.uop_cond_jump(modes, lambda x: x == 0)\n",
    "\n",
    "    def op_lt(self, modes):\n",
    "        return self.uop_cmp(modes, lambda x, y: x < y)\n",
    "\n",
    "    def op_eq(self, modes):\n",
    "        return self.uop_cmp(modes, lambda x, y: x == y)\n",
    "\n",
    "    def op_rebase(self, modes):    \n",
    "        param_mode = self.try_pop_mode(modes)\n",
    "        param_raw = int(self.memory[self.pc + 1])\n",
    "        param = self.uop_read(param_raw, param_mode)\n",
    "\n",
    "        self.offset += param\n",
    "\n",
    "        return self.pc + 2"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 3,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "1863\n"
     ]
    }
   ],
   "source": [
    "raw_code = open(\"puzzle_input/day11.txt\").read()\n",
    "code = raw_code.split(\",\")\n",
    "\n",
    "\n",
    "turn_left = 0\n",
    "turn_right = 1\n",
    "\n",
    "color_black = 0\n",
    "color_white = 1\n",
    "\n",
    "direction_up =    (0,  1)\n",
    "direction_right = (1,  0)\n",
    "direction_down =  (0, -1)\n",
    "direction_left =  (-1, 0)\n",
    "\n",
    "\n",
    "def find_next_direction(current_direction, turn):\n",
    "    if current_direction == direction_up:\n",
    "        if turn == turn_left:\n",
    "            return direction_left\n",
    "        else:\n",
    "            return direction_right\n",
    "        \n",
    "    elif current_direction == direction_right:\n",
    "        if turn == turn_left:\n",
    "            return direction_up\n",
    "        else:\n",
    "            return direction_down\n",
    "        \n",
    "    elif current_direction == direction_down:\n",
    "        if turn == turn_left:\n",
    "            return direction_right\n",
    "        else:\n",
    "            return direction_left\n",
    "        \n",
    "    elif current_direction == direction_left:\n",
    "        if turn == turn_left:\n",
    "            return direction_down\n",
    "        else:\n",
    "            return direction_up\n",
    "    \n",
    "    raise Exception(f\"Could not determine next direction, {current_direction} is not valid\")\n",
    "    \n",
    "assert direction_up == find_next_direction(direction_right, turn_left)\n",
    "assert direction_up == find_next_direction(direction_left, turn_right)\n",
    "assert direction_down == find_next_direction(direction_right, turn_right)\n",
    "assert direction_down == find_next_direction(direction_left, turn_left)\n",
    "assert direction_left == find_next_direction(direction_up, turn_left)\n",
    "assert direction_left == find_next_direction(direction_down, turn_right)\n",
    "assert direction_right == find_next_direction(direction_up, turn_right)\n",
    "assert direction_right == find_next_direction(direction_down, turn_left)\n",
    "\n",
    "def next_position(current_position, current_direction, turn):\n",
    "    current_x, current_y = current_position\n",
    "    next_direction = find_next_direction(current_direction, turn)\n",
    "    \n",
    "    move_x, move_y = next_direction\n",
    "    \n",
    "    next_position = (current_x + move_x, current_y + move_y)\n",
    "    return (next_position, next_direction)\n",
    "\n",
    "\n",
    "assert (-1,  0) == next_position((0, 0), direction_up, turn_left)[0]\n",
    "assert ( 1,  0) == next_position((0, 0), direction_up, turn_right)[0]\n",
    "assert ( 1,  0) == next_position((0, 0), direction_down, turn_left)[0]\n",
    "assert (-1,  0) == next_position((0, 0), direction_down, turn_right)[0]\n",
    "assert ( 0, -1) == next_position((0, 0), direction_left, turn_left)[0]\n",
    "assert ( 0,  1) == next_position((0, 0), direction_left, turn_right)[0]\n",
    "assert ( 0,  1) == next_position((0, 0), direction_right, turn_left)[0]\n",
    "assert ( 0, -1) == next_position((0, 0), direction_right, turn_right)[0]\n",
    "\n",
    "def parse_output_instructions(outputs):\n",
    "    instructions = []\n",
    "    while outputs:\n",
    "        color = outputs.pop(0)\n",
    "        turn = outputs.pop(0)\n",
    "        instructions.append((color, turn))\n",
    "    return instructions\n",
    "\n",
    "\n",
    "def color_at_position(painted_panels, position):\n",
    "    \n",
    "    if position not in painted_panels:\n",
    "        return color_black\n",
    "    \n",
    "    return painted_panels[position]\n",
    "    \n",
    "\n",
    "def solve(starting_color):\n",
    "\n",
    "    cpu = IntCodeCpu(code)\n",
    "    \n",
    "    input_buffer = [starting_color]\n",
    "    output_buffer = []\n",
    "\n",
    "    cpu.start(input_buffer, output_buffer)\n",
    "\n",
    "    current_direction = direction_up\n",
    "    current_position = (0, 0)\n",
    "\n",
    "    painted_panels = {}\n",
    "\n",
    "    while cpu.stalled:\n",
    "        output_instructions = parse_output_instructions(output_buffer)\n",
    "\n",
    "        for color, turn in output_instructions:\n",
    "            painted_panels[current_position] = color\n",
    "            (current_position, current_direction) = next_position(current_position, current_direction, turn)\n",
    "\n",
    "        input_buffer.append(color_at_position(painted_panels, current_position))\n",
    "\n",
    "        cpu.run()\n",
    "        \n",
    "    return painted_panels\n",
    "    \n",
    "print(len(solve(color_black)))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 9,
   "metadata": {},
   "outputs": [
    {
     "data": {
      "text/html": [
       "<img src=\"./img/day11.png\"/>"
      ],
      "text/plain": [
       "<IPython.core.display.Image object>"
      ]
     },
     "metadata": {},
     "output_type": "display_data"
    }
   ],
   "source": [
    "from PIL import Image, ImageDraw\n",
    "from IPython.display import Image as IPythonImage\n",
    "from IPython.display import display\n",
    "\n",
    "panels = solve(color_white)\n",
    "\n",
    "min_x = min(x for (x,y) in panels if panels[(x,y)])\n",
    "max_x = max(x for (x,y) in panels if panels[(x,y)])\n",
    "min_y = min(y for (x,y) in panels if panels[(x,y)])\n",
    "max_y = max(y for (x,y) in panels if panels[(x,y)])\n",
    "\n",
    "offset_x = 0 - min_x\n",
    "offset_y = 0 - min_y\n",
    "\n",
    "img = Image.new(\"RGB\", (max_y + offset_y + 20, max_x + offset_x + 20), \"black\")\n",
    "\n",
    "draw = ImageDraw.Draw(img)\n",
    "\n",
    "color_lookup = {\n",
    "    0: \"black\",\n",
    "    1: \"white\"\n",
    "}\n",
    "                \n",
    "for point in panels:\n",
    "    x, y = point\n",
    "    color = panels[point]\n",
    "    draw.point((y + offset_y + 10, x + offset_x + 10), fill=color_lookup[color])\n",
    "    \n",
    "img.save('img/day11.png')\n",
    "\n",
    "display(IPythonImage(url=\"./img/day11.png\"))    "
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.9.0"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
